/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package FunctionLayer

import DataLayer.DataMapper
import FunctionLayer.StanfordParser.SentimentAnalyzerTest
import com.google.common.base.Stopwatch
import edu.stanford.nlp.ie.AbstractSequenceClassifier
import edu.stanford.nlp.ie.crf.CRFClassifier
import edu.stanford.nlp.ling.CoreAnnotations
import edu.stanford.nlp.ling.CoreLabel
import edu.stanford.nlp.ling.TaggedWord
import edu.stanford.nlp.parser.lexparser.LexicalizedParser
import edu.stanford.nlp.pipeline.Annotation
import edu.stanford.nlp.pipeline.CoreDocument
import edu.stanford.nlp.pipeline.StanfordCoreNLP
import edu.stanford.nlp.tagger.maxent.MaxentTagger
import edu.stanford.nlp.trees.GrammaticalStructureFactory
import edu.stanford.nlp.trees.Tree
import edu.stanford.nlp.util.CoreMap
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.yield
import java.util.*
import java.util.concurrent.TimeUnit
import java.util.regex.Pattern
import kotlin.collections.ArrayList
import kotlin.collections.HashMap


/**
 *
 * @author install1
 */
public class Datahandler {
    private val stopwatch: Stopwatch
    private val EXPIRE_TIME_IN_MINUTES = TimeUnit.MINUTES.convert(30, TimeUnit.MINUTES)
    private var pipelineAnnotationCache: HashMap<String, Annotation>
    private var pipelineSentimentAnnotationCache = HashMap<String, Annotation>()
    private var coreDocumentAnnotationCache: HashMap<String, CoreDocument>
    private var jmweAnnotationCache = HashMap<String, Annotation>()
    private val stringCache = ArrayList<String>()
    private val nerModel = "edu/stanford/nlp/models/ner/english.all.3class.caseless.distsim.crf.ser.gz"
    private var tagger: MaxentTagger = MaxentTagger()
    private var gsf: GrammaticalStructureFactory
    private var classifier: AbstractSequenceClassifier<CoreLabel>

    //SentimentAnalyzer Hashmaps
    private var tokenizeCountingHashMap: HashMap<String, Int> = HashMap()
    private var taggedWordListHashMap: HashMap<String, List<List<TaggedWord>>> = HashMap()
    private var retrieveTGWListHashMap: HashMap<String, java.util.ArrayList<String>> =
            HashMap()
    private var sentences1HashMap: HashMap<String, List<CoreMap>> = HashMap()
    private var sentencesSentimentHashMap: HashMap<String, List<CoreMap>> = HashMap()
    private var trees1HashMap: HashMap<String, java.util.ArrayList<Tree>> = HashMap()

    constructor() {
        stopwatch = Stopwatch.createUnstarted()
        jmweAnnotationCache = HashMap<String, Annotation>()
        pipelineAnnotationCache = HashMap<String, Annotation>()
        pipelineSentimentAnnotationCache = HashMap<String, Annotation>()
        coreDocumentAnnotationCache = HashMap<String, CoreDocument>()
        gsf = initiateGrammaticalStructureFactory()
        classifier = CRFClassifier.getClassifierNoExceptions(nerModel)
    }

    fun initiateGrammaticalStructureFactory(): GrammaticalStructureFactory {
        val options = arrayOf("-maxLength", "100")
        val lexParserEnglishRNN = "edu/stanford/nlp/models/lexparser/englishRNN.ser.gz"
        val lp = LexicalizedParser.loadModel(lexParserEnglishRNN, *options)
        val tlp = lp.getOp().langpack()
        return tlp.grammaticalStructureFactory()
    }

    public fun pipeLineSetUp(): StanfordCoreNLP {
        val props = Properties()
        val shiftReduceParserPath = "edu/stanford/nlp/models/srparser/englishSR.ser.gz"
        val nerModel2 = "edu/stanford/nlp/models/ner/english.conll.4class.caseless.distsim.crf.ser.gz"
        val nerModel3 = "edu/stanford/nlp/models/ner/english.muc.7class.caseless.distsim.crf.ser.gz"
        props.setProperty("annotators", "tokenize,ssplit,pos,lemma,ner,parse")
        props.setProperty("parse.model", shiftReduceParserPath)
        props.setProperty("parse.maxlen", "90")
        props.setProperty("parse.binaryTrees", "true")
        props.setProperty("threads", "5")
        props.setProperty("pos.maxlen", "90")
        props.setProperty("tokenize.maxlen", "90")
        props.setProperty("ssplit.maxlen", "90")
        props.setProperty("lemma.maxlen", "90")
        props.setProperty("ner.model", "$nerModel,$nerModel2,$nerModel3")
        props.setProperty("ner.combinationMode", "HIGH_RECALL")
        props.setProperty("regexner.ignorecase", "true")
        props.setProperty("ner.fine.regexner.ignorecase", "true")
        props.setProperty("tokenize.options", "untokenizable=firstKeep")
        return StanfordCoreNLP(props)
    }

    fun shiftReduceParserInitiate(): StanfordCoreNLP {
        val propsSentiment = Properties()
        val lexParserEnglishRNN = "edu/stanford/nlp/models/lexparser/englishRNN.ser.gz"
        val sentimentModel = "edu/stanford/nlp/models/sentiment/sentiment.ser.gz"
        val taggerPath = "edu/stanford/nlp/models/pos-tagger/english-left3words/english-left3words-distsim.tagger"
        val customStopWordList = "start,starts,period,periods,a,an,and,are,as,at,be,but,by,for,if,in,into,is,it,no,not,of," +
                "on,or,such,that,the,their,then,there,these,they,this,to,was,will,with"
        propsSentiment.setProperty("parse.model", lexParserEnglishRNN)
        propsSentiment.setProperty("sentiment.model", sentimentModel)
        propsSentiment.setProperty("parse.maxlen", "90")
        propsSentiment.setProperty("threads", "5")
        propsSentiment.setProperty("pos.maxlen", "90")
        propsSentiment.setProperty("tokenize.maxlen", "90")
        propsSentiment.setProperty("ssplit.maxlen", "90")
        propsSentiment.setProperty("annotators", "tokenize,ssplit,pos,parse,sentiment,lemma,stopword") //coref too expensive memorywise
        propsSentiment.setProperty("customAnnotatorClass.stopword", "FunctionLayer.StopwordAnnotator")
        propsSentiment.setProperty(StopwordAnnotator.STOPWORDS_LIST, customStopWordList)
        propsSentiment.setProperty("tokenize.options", "untokenizable=firstKeep")
        tagger = MaxentTagger(taggerPath)

        println("finished shiftReduceParserInitiate\n")
        return StanfordCoreNLP(propsSentiment)
    }

    fun updateStringCache() {
        if (stopwatch.elapsed(TimeUnit.MINUTES) >= EXPIRE_TIME_IN_MINUTES || !stopwatch.isRunning) {
            if (!stopwatch.isRunning) {
                stopwatch.start()
            } else {
                stopwatch.reset()
            }
            stringCache.sortWith(Comparator.comparingInt(String::length).reversed());
            System.out.println("pre InsertMYSQLStrings")
            DataMapper.InsertMYSQLStrings(stringCache)
        }
    }

    fun initiateMYSQL() {
        stringCache.addAll(DataMapper.getAllStrings())
    }

    private fun trimString(str: String): String {
        var message = str.trim { it <= ' ' }
        if (message.startsWith("<@")) {
            message = message.substring(message.indexOf("> ") + 2)
        }
        if (!message.isEmpty()) {
            message = message.replace("@", "")
            if (message.contains("<>")) {
                message = message.substring(message.indexOf(">"))
            }
            if (message.startsWith("[ *")) {
                message = message.substring(message.indexOf("]"))
            }
        }
        return message
    }

    private fun createStrAnnotation(str: String, stanfordCoreNLP: StanfordCoreNLP, sentimentBool: Boolean) {
        val strAnno2 = Annotation(str)
        strAnno2.compact()
        stanfordCoreNLP.annotate(strAnno2)
        if (sentimentBool) {
            pipelineSentimentAnnotationCache.put(str, strAnno2)
        } else {
            pipelineAnnotationCache.put(str, strAnno2)
        }
    }

    private fun getResponseFutures(strF: String, stanfordCoreNLP: StanfordCoreNLP, stanfordCoreNLPSentiment: StanfordCoreNLP): String {
        val strAnno: Annotation = Annotation(strF)
        strAnno.compact()
        stanfordCoreNLP.annotate(strAnno)

        val strAnnoSentiment: Annotation = Annotation(strF)
        strAnnoSentiment.compact()
        stanfordCoreNLPSentiment.annotate(strAnnoSentiment)

        val coreDocument = CoreDocument(strF)
        stanfordCoreNLP.annotate(coreDocument)

        val values_copy: List<String> = ArrayList(stringCache)
        var preRelationUserCounters = -155000.0
        val concurrentRelations: MutableList<String> = arrayListOf()
        val SB = StringBuilder()
        var jmweAnnotationF = PipelineJMWESingleton.INSTANCE.getJMWEAnnotation(strF)
        var tokenizeCountingF: Int? = null
        var taggedWordListF: List<List<TaggedWord>>? = null
        var retrieveTGWListF: java.util.ArrayList<String>? = null
        var sentencesF: List<CoreMap>? = null
        var sentencesSentimentF: List<CoreMap>? = null
        var coreMaps1: List<CoreMap> = jmweAnnotationF.get(CoreAnnotations.SentencesAnnotation::class.java)
        var treesF: java.util.ArrayList<Tree>? = null


        for (str1 in values_copy) {
            if (strF != str1) {
                val annotation2 = pipelineSentimentAnnotationCache.getOrDefault(str1, null)
                val annotation4 = pipelineAnnotationCache.getOrDefault(str1, null)
                val coreDocument1 = coreDocumentAnnotationCache.getOrDefault(str1, null)
                var jmweAnnotation = jmweAnnotationCache.getOrDefault(str1, null)
                if (annotation2 == null) {
                    createStrAnnotation(str1, stanfordCoreNLPSentiment, true)
                }
                if (annotation4 == null) {
                    createStrAnnotation(str1, stanfordCoreNLP, false)
                }
                if (coreDocument1 == null) {
                    getCoreDocumentsSuggested(stanfordCoreNLP, str1)
                }
                if (jmweAnnotation == null) {
                    getJMWEAnnotation(str1)
                    jmweAnnotation = jmweAnnotationCache.get(str1)
                }
                val tokenizeCounting: Int? = tokenizeCountingHashMap.getOrDefault(str1, null)
                val taggedWordList1: List<List<TaggedWord>>? = taggedWordListHashMap.getOrDefault(str1, null)
                val retrieveTGWList1: java.util.ArrayList<String>? = retrieveTGWListHashMap.getOrDefault(str1, null)
                val sentence1: List<CoreMap>? = sentences1HashMap.getOrDefault(str1, null)
                val sentenceSentiment1: List<CoreMap>? = sentencesSentimentHashMap.getOrDefault(str1, null)
                val trees1 = trees1HashMap.getOrDefault(str1, null)
                var coreMaps2: List<CoreMap> = listOf()
                if (jmweAnnotation != null) {
                    coreMaps2 = jmweAnnotation.get(CoreAnnotations.SentencesAnnotation::class.java)
                }
                var SMX = SentimentAnalyzerTest(strF, str1, SimilarityMatrix(strF, str1),
                        coreMaps1, coreMaps2, strAnno,
                        pipelineAnnotationCache[str1], strAnnoSentiment,
                        pipelineSentimentAnnotationCache[str1], coreDocument, coreDocumentAnnotationCache[str1],
                        tagger, gsf, classifier, tokenizeCounting, tokenizeCountingF,
                        taggedWordListF, taggedWordList1, retrieveTGWListF, retrieveTGWList1,
                        sentencesF, sentence1, sentencesSentimentF, sentenceSentiment1, treesF, trees1)
                if (tokenizeCounting == null) {
                    val tokenizeCounting1 = SMX.getTokenizeCounting();
                    tokenizeCountingHashMap.put(str1, tokenizeCounting1)
                }
                if (taggedWordList1 == null) {
                    val taggedWordList1Local = SMX.getTaggedWordList1();
                    taggedWordListHashMap.put(str1, taggedWordList1Local)
                }
                if (tokenizeCountingF == null) {
                    val tokenizeCountingF1 = SMX.getTokenizeCountingF();
                    tokenizeCountingF = tokenizeCountingF1;
                }
                if (taggedWordListF == null) {
                    val taggedWordListF1 = SMX.getTaggedWordListF();
                    taggedWordListF = taggedWordListF1;
                }
                if (retrieveTGWListF == null) {
                    val retrieveTGWListF1 = SMX.getRetrieveTGWListF();
                    retrieveTGWListF = retrieveTGWListF1;
                }
                if (retrieveTGWList1 == null) {
                    val retrieveTGWList11 = SMX.getRetrieveTGWList1();
                    retrieveTGWListHashMap.put(str1, retrieveTGWList11);
                }
                if (sentencesF == null) {
                    val sentencesF1 = SMX.getSentencesF();
                    sentencesF = sentencesF1;
                }
                if (sentence1 == null) {
                    val sentences1 = SMX.getSentences1();
                    sentences1HashMap.put(str1, sentences1)
                }
                if (sentencesSentimentF == null) {
                    val sentencesSentimentF1 = SMX.getSentencesSentimentF();
                    sentencesSentimentF = sentencesSentimentF1;
                }
                if (sentenceSentiment1 == null) {
                    val sentencesSentiment1 = SMX.getSentencesSentiment1();
                    sentencesSentimentHashMap.put(str1, sentencesSentiment1);
                }
                if (treesF == null) {
                    val treesF1 = SMX.getTreesF();
                    treesF = treesF1;
                }
                if (trees1 == null) {
                    val trees11 = SMX.getTrees1();
                    trees1HashMap.put(str1, trees11)
                }

                var getSMX: SimilarityMatrix = SMX.callSMX()
                val scoreRelationLastUserMsg = getSMX.distance
                if (scoreRelationLastUserMsg > preRelationUserCounters) {
                    preRelationUserCounters = scoreRelationLastUserMsg
                    concurrentRelations.add(getSMX.secondaryString)
                }
            }
        }
        val cacheRequirement = 6500;
        if (preRelationUserCounters > cacheRequirement && !stringCache.contains(strF) && filterContent(strF)) {
            stringCache.add(strF)
        }
        val randomLenghtPermit = strF.length * (Math.random() * Math.random() * Math.random() * (Math.random() * 10))
        Collections.reverse(concurrentRelations)
        if (!concurrentRelations.isEmpty()) {
            for (secondaryRelation in concurrentRelations) {
                if (SB.toString().length > randomLenghtPermit && !SB.toString().isEmpty()) {
                    break
                }
                SB.append(secondaryRelation).append(" ")
            }
        }
        if (SB.toString().isEmpty()) {
            return "failure, preventing stuckness"
        }
        return SB.toString()
    }

    private fun getJMWEAnnotation(str1: String) {
        val jmweAnnotation = PipelineJMWESingleton.INSTANCE.getJMWEAnnotation(str1)
        jmweAnnotationCache.put(str1, jmweAnnotation)
    }

    fun getResponseMsg(str: String, personName: String, stanfordCoreNLP: StanfordCoreNLP,
                       stanfordCoreNLPSentiment: StanfordCoreNLP, ingameResponse: Boolean): String {
        var responseFutures: String = ""
        runBlocking {
            val launch1 = launch(Dispatchers.Default) {
                var strF = trimString(str)
                responseFutures = getResponseFutures(strF, stanfordCoreNLP, stanfordCoreNLPSentiment)
                if (!ingameResponse) {
                    responseFutures = checkPersonPresentInSentence(personName, responseFutures, strF, stanfordCoreNLP,
                            stanfordCoreNLPSentiment)
                }
                yield()
            }
            launch1.join()
        }
        return responseFutures
    }

    private fun checkPersonPresentInSentence(personName: String, responseMsg: String, userLastMessage: String,
                                             stanfordCoreNLP: StanfordCoreNLP,
                                             stanfordCoreNLPSentiment: StanfordCoreNLP): String {
        try {
            val pipelineCoreDcoument = CoreDocument(responseMsg)
            val pipelineCoreDcoumentLastMsg = CoreDocument(userLastMessage)
            stanfordCoreNLP.annotate(pipelineCoreDcoument)
            stanfordCoreNLPSentiment.annotate(pipelineCoreDcoumentLastMsg)
            val regex = "(.*?\\d){10,}"
            for (em in pipelineCoreDcoument.entityMentions()) {
                val entityType = em.entityType()
                if (entityType == "PERSON") {
                    var str = responseMsg
                    val emText = em.text()
                    val pattern = Pattern.compile(regex)
                    val matcher = pattern.matcher(personName)
                    val isMatched = matcher.matches()
                    if (emText != personName && !isMatched) {
                        for (emLastMsg in pipelineCoreDcoumentLastMsg.entityMentions()) {
                            if (emText != emLastMsg.text() && !Character.isDigit(emLastMsg.text().trim { it <= ' ' }[0])) {
                                //System.out.println("emLastMsg.text(): " + emLastMsg.text());
                                str = (responseMsg.substring(0, responseMsg.indexOf(emText)) + " "
                                        + emLastMsg + " " + responseMsg.substring(responseMsg.indexOf(emText)))
                            }
                        }
                        str += " $personName"
                        return str
                    }
                }
            }
        } catch (e: Exception) {
            println("""SCUFFED JAYZ: ${e.localizedMessage}""".trimIndent())
        }
        return responseMsg
    }

    fun filterContent(str: String): Boolean {
        if (!str.isEmpty() && str.length > 3) {
            var str1Local: String = str.trim();
            if (str1Local.length > 2 && !str1Local.startsWith("!")) {
                return true
            }
        }
        return false
    }

    fun getCoreDocumentsSuggested(pipeline: StanfordCoreNLP, str: String) {
        val annotation = Annotation(str)
        pipeline.annotate(annotation)
        val coreDocument = CoreDocument(annotation)
        coreDocumentAnnotationCache.put(str, coreDocument)
    }
}